// #mark Error types

var DeliveryError = {
	notFound: 'notFound',
	connection: 'connection',
	server: 'server',
	signIn: 'signIn',
	password: 'password',
	cookies: 'cookies',
	unsupported: 'unsupported',
	unknown: 'unknown'
}

// #mark User agent

function userAgentForService (service) {
	if (service.desktopUserAgent) return 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_3) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/12.0.3 Safari/605.1.15';
	if (service.mobileUserAgent) return 'Mozilla/5.0 (iPhone; CPU iPhone OS 12_1_4 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/12.0 Mobile/15E148 Safari/604.1';
	return null;
}

// #mark Update queue

function Queue () {

	this.reset = function () {
		this.pendingItems = [];
		this.activeItems = [];
		this.froms = [];
		this.active = [];
		this.delay = 500;
	}
	
	this.reset();

	this.add = function (delivery) {
		if (delivery.passRequired && !delivery.oneByOne) {
			var checkValue = delivery.extra;
			var previousValue = this.froms[delivery.from];
			if (previousValue) {
				if (previousValue != checkValue) delivery.oneByOne = true;
			} else {
				this.froms[delivery.from] = checkValue;
			}
		}
		this.pendingItems.push(delivery);
	}

	this.update = function () {
		if (this.pendingItems.length) {

			var addBack = [];

			do {
				var delivery = this.pendingItems.shift();
				var oneByOne = delivery.oneByOne;
				if (oneByOne && this.active[delivery.from]) {
					addBack.push(delivery);
				} else {
					this.activeItems.push(delivery);
					this.active[delivery.from] = true;
					delivery.updateStatus();
				}
			} while (this.pendingItems.length);

			while (addBack.length) {
				var delivery = addBack.shift();
				this.pendingItems.push(delivery);
			}

		} else if (dashboardWidget) {
			setTimeout(finishUpdate,500);
		}
	}

	this.completed = function (delivery) {
		var index = this.activeItems.indexOf(delivery);
		if (index >= 0) this.activeItems.splice(index,1);
		this.active[delivery.from] = false;
		if (!this.busy()) this.reset();
	}

	this.busy = function () {
		if (this.pendingItems.length > 0) return true;
		if (this.activeItems.length > 0) return true;
		return false;
	}

}

// #mark Deliveries and updating

function Service (from, no) {
	if (from) {
		this.no = no || '';
		this.extra = '';
		this.setService(from);
	}
}

function Delivery (settings) {
	Service.call(this);
	if (dashboardWidget) {
		this.loading = false;
		this.spinAngle = 0;
	}
	this.setSettings(settings);
}

Delivery.prototype = Object.create(Service.prototype);

Delivery.prototype.constructor = Service;

Delivery.prototype.setSettings = function (settings) {

	this.req = [];
	this.reqCount = 0;
	this.reqComplete = 0;

	//  Success values:
	// -1: Couldn't connect: no internet or keychain locked
	//  0: Couldn't connect: site down
	//  1: Not found, no password, needs sign in, or no cookies
	//  2: Found

	this.success = 2;
	this.lastUpdate = false;
	this.lastError = false;
	this.nextUpdate = 0;
	this.backgroundUpdate = 0;
	this.interval = 0;
	this.items = [];
	this.previousItems = [];
	this.count = 0;

	this.fromEdited = false;
	this.skipExtra = false;
	this.tries = 0;
	this.tap = false;

	// Initialization
	this.id = settings.uuid;
	this.hash = settings.hash || '';
	this.name = settings.name || '';
	this.extra = settings.extra || '';
	this.no = settings.no || '';
	this.date = dateFromISOString(settings.date);
	this.modified = (settings.modified) ? parseFloat(settings.modified) : 0;
	this.displayed = settings.displayed || 0;

	this.usedItemCache = false;
	this.clearItemCache = false;
	var cacheJSON = settings.itemCacheJSON;
	// Make sure the cache includes the number, in case the delivery was edited
	if (cacheJSON && cacheJSON.indexOf(this.no) !== -1) {
		this.itemCache = JSON.parse(cacheJSON);
	}

	this.setService(settings.from);
	this.fromName = getFromName(this.from,false,false);
	this.fromTrans = getFromName(this.from,true,false);

	if (dashboardWidget) {
		this.lastUpdated = (settings.lastupdated) ? parseFloat(settings.lastupdated) : 0;
		this.fromClass = getClassName(this.from);
	}

}

Delivery.prototype.updateStatus = function () {
	if (dashboardWidget) this.startLoad();
	this.resetRequests(true);
	this.success = 2;
	showMessage(this, translateFormat('Contacting %@…', this.fromTrans));
	this.previousItems = this.items;
	this.startUpdate(this,false);
}

Delivery.prototype.loadItemCache = function (method, site) {

	this.items = [];

	for (var x in this.itemCache) {

		var cache = this.itemCache[x];
		var item = new DeliveryPackage();
		item.title = cache.itemName;

		item.status = '';
		item.details = {};
		this.items[x] = item;

		if (cache.url) {
			var options = new RequestOptions(site, x);
			this.loadPage(cache.url, method, options);
		}

	}

	this.count = this.items.length;
	this.usedItemCache = true;
	this.updateItems();

}

function RequestOptions (site, childNo) {
	this.site = site;
	this.childNo = (childNo === false) ? 0 : childNo;
	this.isChild = (childNo === false) ? false : true;
}

function DummyRequest (delivery, responseText, options) {
	this.delivery = delivery;
	this.options = options;
	this.responseText = responseText;
	this.status = 200;
	this.readyState = 4;
	this.abort = function(){};
	delivery.getTracker(this);
}

Delivery.prototype.loadSection = function (section, options) {
	var reqNo = this.reqCount;
	this.reqCount++;
	this.req[reqNo] = new DummyRequest(this, section, options);
}

Delivery.prototype.startRequests = function () {

	var nextRequest = null;
	var busy = false;

	for (var i in this.req) {
		var request = this.req[i];
		if (request) {
			var state = request.readyState;
			if (!state) {
				if (!nextRequest) nextRequest = request;
			} else if (state < 4) {
				busy = true;
			}
		}
	}

	var started = false;
	if (!busy && nextRequest) {
		nextRequest.startRequest();
		started = true;
	}

	return started;

}

Delivery.prototype.resetRequests = function (cancelAll) {
	if (cancelAll) {
		for (var i in this.req) {
			var request = this.req[i];
			if (request) {
				if (typeof request.abort == 'function') request.abort();
				if (request.timer) clearTimeout(request.timer);
			}
		}
	}
	this.reqCount = 0;
	this.reqComplete = 0;
}

Delivery.prototype.connectFailed = function (connectionDown) {
	if (!this.count) this.count = -1;
	this.notFound(false, connectionDown);
	this.reqComplete++;
	this.updateItems();
}

Delivery.prototype.notFound = function (message, connectionDown, showImmediately) {

	if (dashboardWidget) this.stopLoad();
	if (this.count == 0) {

		if (manualSignIn[this.from+'-'+this.extra]) {

			var item = errorItem(this);
			item.shippedText = translate('Sign in required');
			if (dashboardWidget) {
				item.deliveredText = '<a href="javascript:openSafariSite(\''+this.url+'\',\''+this.from+'\',\''+this.extra+'\');">'+
					translate('Click to sign in.')+'</a>';
				item.clickable = true;
			} else {
				item.deliveredText = translate('Please sign in.');
			}
			this.tap = 'signin';
			this.success = 1;
			this.items = [item];
			this.count = -1;
			this.displayed = 0;

		} else if (dashboardWidget && DeliveryStatus && !DeliveryStatus.cookiesEnabled()) {

			var item = errorItem(this);
			item.shippedText = translate('Cookies are required');
			item.deliveredText = '<a href="javascript:openSite(\'https://junecloud.com/sync/help/delivery-status-widget/cookies.html\');">'+
				translate('Click here for help.')+'</a>';
			item.clickable = true;
			this.success = 1;
			this.items = [item];
			this.count = -1;
			this.displayed = 0;

		} else if (showImmediately && message) {

			var item = errorItem(this);
			item.shippedText = message;
			item.deliveredText = translate('Please check your settings.');
			this.success = 2;
			this.items = [item];
			this.count = -1;
			this.displayed = 0;		

		} else if (!dashboardWidget || this.shouldShowError()) { 

			var item = errorItem(this);
			item.shippedText = (message) ? message : translate('Not found');
			item.deliveredText = translate('Please check your settings.');
			this.success = 1;
			this.items = [item];
			this.count = -1;
			this.displayed = 0;

		} else {

			// Ideally, we should stop setting the count in Services.js
			// Just pass the type of error to this function, so it can set it!
			logMessage(1,'Ignoring not found error for '+this.from+'. Last success: '+this.lastUpdate);
			this.items = this.previousItems;
			this.count = this.items.length;
			this.success = 0;
			if (!this.lastError) this.lastError = new Date();

		}

	} else if (this.count == -1) {

		var item = errorItem(this);
		item.shippedText = translate('Couldn’t connect');
		item.deliveredText = translate('Please try again later.');

		this.items = [item];
		this.displayed = 0;
		this.success = (connectionDown) ? -1 : 0;

	} else {

		this.items = this.previousItems;
		this.success = (connectionDown) ? -1 : 0;

	}

	if (dashboardWidget) checkVersion(3,true);

}

Delivery.prototype.notSupported = function (isChild, isUnknown) {

	if (isChild) return;

	var item = errorItem(this);
	item.icon = 'transit';

	if (this.url) {
		item.shippedText = translate('This shipment can only be viewed online.');
		if (dashboardWidget) {
			item.deliveredText = translate('Click to view this delivery.');
		} else {
			item.deliveredText = (typeof usesClicks !== 'undefined' && usesClicks) ? 
				translate('Click “view details online” to show it.') : translate('Tap “view details online” to show it.');
		}
	} else {
		item.shippedText = translate('This service is not supported.');
		item.deliveredText = translate('Delivery date unknown');
	}
	item.shippedBold = true;
	item.deliveredBold = false;

	this.items = [item];
	this.count = 1;
	this.displayed = 0;
	this.success = 2;

}

Delivery.prototype.shouldShowError = function () {
	if (!this.lastUpdate) return true;
	if (this.previousItems.length == 0) return true;
	if (this.previousItems[0] && this.previousItems[0].icon == 'error') return true;
	if (!this.lastError) return false; // First error
	if ((this.lastError.getTime()/1000) + 18000 < getUnixTime()) return true; // 5 hours
	return false;
}

Delivery.prototype.updatePlaceholder = function () {

	var item = new DeliveryPackage();
	if (this.name) {
		item.title = htmlEntities(this.name);
	} else if (this.no) {
		item.title = htmlEntities(this.no);
	} else if (this.extra) {
		var domain = this.extra.replace(/^[a-z]+:\/\//i,'');
		domain = domain.replace(/\/.*/i,'');
		item.title = domain;
	}

	if (this.url) {
		item.shippedText = translate('This shipment can only be viewed online.');
		item.shippedBold = false;
	} else {
		item.shippedText = translate('Ship date unknown');
		item.shippedBold = false;
	}

	var date = this.date;
	if (date) {
		var daysApart = daysUntilDate(date);
		var dateString = date.toISOString();
		item.icon = (daysApart < 0) ? 'checkmark' : '';
		item.dateString = dateString;
		item.deliveredText = translate('Delivered by %@');
		item.deliveredBold = false;
		item.deliveredDateString = dateString;
	} else {
		item.icon = 'transit';
		item.dateString = '';
		item.deliveredText = translate('Delivery date unknown');
		item.deliveredBold = false;
		item.deliveredDateString = '';
	}
	this.items = [];
	this.items[0] = item;
	this.count = 1;

}

Delivery.prototype.updateFailed = function (error, message) {
	if (error == DeliveryError.server && this.count == 0) this.count = -1;
	this.notFound(message, false, false);
}

Delivery.prototype.updateComplete = function () {
	return (this.reqComplete >= this.reqCount) ? true : false;
}

Delivery.prototype.updateItems = function () {

	if (this.updateComplete() == false) {
		this.startRequests();
		return;
	}

	if (this.clearItemCache && this.usedItemCache) {
		this.itemCache = null;
		this.usedItemCache = false;
		this.clearItemCache = false;
		this.updateStatus();
		return;
	}

	this.req = [];
	this.previousItems = [];

	var newInterval = 99999999;
	switch (this.success) {
		case -1: newInterval = 0; break;
		case 0: newInterval = 60; break;
		case 1: newInterval = (this.tap == 'password' || this.tap == 'signin') ? 1440 : 60; break;
		default: break;
	}

	if (this.from != 'other' && this.count > 0) {

		for (var x in this.items) {

			var item = this.items[x];
			var hasDeliveryDays = false;

			var shippedDate = item.shippedDate;
			var deliveredDate = item.deliveredDate;
			var isComplete = item.isComplete;
			var statusDateString = null;

			if (shippedDate || !item.shippedText) {

				var isShipped = (isComplete || item.status.match(/^shipped$/i)) ? true : false;
				item.shippedText = translate('Ship date unknown');
				item.shippedBold = false;

				if (isComplete) {

					if (item.status) {
						item.shippedText = item.status;
					} else {
						item.shippedText = translate('Delivered!');
					}
					item.shippedBold = true;

				} else if (shippedDate) {

					if (isShipped) {
						item.shippedText = translate('Shipped!');
						item.shippedExtra = translate('Shipped %@');
						item.shippedBold = true;
					} else item.shippedText = translate('Shipped by %@');
					if (this.countToShip) item.daysNo = daysUntilDate(shippedDate);

				} else if (isShipped) {

					item.shippedText = translate('Shipped!');
					item.shippedBold = true;

				}

				if (item.details && item.details.status) {
					var status = stripHTML(item.details.status);
					var location = stripHTML(item.details.location);
					if (location || (status && !status.match(/^delivered$/i))) {

						if (shippedDate && item.shippedText) item.shippedExtra = item.shippedText;
						item.shippedText = status;
						if (location && item.shippedText.indexOf(location) == -1) item.shippedText += ': '+location;

						var detailDateString = item.details.dateString;
						if (detailDateString) {
							item.shippedExtra = translate('Status updated %@');
							statusDateString = detailDateString;
						}
						item.shippedBold = true;

					}
				} else if (item.status && !isComplete && !isShipped) {
					item.shippedText = item.status;
					item.shippedBold = true;
				}

				var dateString = (shippedDate) ? shippedDate.toISOString() : '';
				item.shippedDateString = dateString;
				if (this.countToShip) item.dateString = dateString;

			}

			if (deliveredDate || !item.deliveredText) {

				var deliverDaysNo = false;
				item.deliveredText = translate('Delivery date unknown');
				item.deliveredBold = false;

				if (isComplete) {

					item.deliveredText = translate('Delivered!');
					item.deliveredBold = true;
					if (deliveredDate) {
						if (item.shippedText == item.deliveredText) {
							item.deliveredText = translate('Delivered %@');
							item.deliveredBold = false;
						} else item.deliveredExtra = translate('Delivered %@');
					}
					deliverDaysNo = daysUntilDate(deliveredDate);

				} else if (deliveredDate) {

					item.deliveredText = translate('Delivered by %@');
					deliverDaysNo = daysUntilDate(deliveredDate);

				}

				if (deliverDaysNo !== false) {
					item.daysNo = deliverDaysNo;
					hasDeliveryDays = true;
				}

				var dateString = (deliveredDate) ? deliveredDate.toISOString() : '';
				item.deliveredDateString = dateString;
				if (dateString) item.dateString = dateString;

			}

			if (isComplete) {
				item.icon = 'checkmark';
			} else if (item.daysNo > 0 || (item.daysNo === 0 && hasDeliveryDays)) {
				item.icon = '';
			} else if (!item.icon) {
				item.icon = 'transit';
			}

			var minutes = 40;
			if (this.neverUpdate) {
				minutes = 99999999;
			} else if (item.icon == 'checkmark') {
				var daysModified = (this.modified) ? daysUntilDate(this.modified*1000) : 0;
				minutes = (daysModified <= -7) ? 99999999 : 1440;
			} else if (item.icon == 'error') {
				if (this.tap == 'password' || this.tap == 'signin') {
					minutes = 99999999;
				} else {
					var daysModified = (this.modified) ? daysUntilDate(this.modified*1000) : 0;
					if (daysModified <= -30) {
						minutes = 99999999;
					} else if (daysModified <= -10) {
						minutes = 1440;
					} else if (daysModified <= -3) {
						minutes = 180;
					} else {
						minutes = 60;
					}
				}
			} else if (item.daysNo === false) {
				var daysModified = 0;
				if (statusDateString) {
					var statusDate = dateFromISOString(statusDateString);
					daysModified = daysUntilDate(statusDate);
				} else if (this.modified) {
					daysModified = daysUntilDate(this.modified*1000);
				}
				if (daysModified <= -30) {
					minutes = 1440;
				} else if (daysModified <= -15) {
					minutes = 360;
				} else if (daysModified <= -5) {
					minutes = 60;
				}
			} else if (item.daysNo <= -1) {
				var daysApart = (hasDeliveryDays) ? item.daysNo : -1;
				if (daysApart <= -30) {
					minutes = 1440;
				} else if (daysApart <= -15) {
					minutes = 720;
				} else if (daysApart <= -5) {
					minutes = 180;
				}
			} else if (item.daysNo <= 3) {
				minutes = 20;
			} else if (item.daysNo <= 10) {
				minutes = 60;
			} else if (item.daysNo <= 30) {
				minutes = 360;
			} else if (item.daysNo > 30) {
				minutes = 1440;
			}
			if (minutes < newInterval) newInterval = minutes;
			if (statusDateString) item.statusDateString = statusDateString;

			this.items[x] = item;

		}

	}

	if (this.success > 0) {
		this.lastUpdate = new Date();
		this.lastError = false;
	}
	this.resetRequests(false);
	this.interval = newInterval;

	if (dashboardWidget) {
		var rightNow = getUnixTime();
		this.nextUpdate = rightNow + (newInterval * 60);
		var backgroundInterval = newInterval;
		if (backgroundInterval < 1440) {
			backgroundInterval = backgroundInterval * 3;
			if (backgroundInterval >= 720) backgroundInterval = Math.max(newInterval,720);
			if (backgroundInterval < 30) backgroundInterval = 30;
		}
		this.backgroundUpdate = rightNow + (backgroundInterval * 60);
		if (this.success > 0) {
			hadSuccess = true;
			if (this.success == 2) setHistoryTitle(this.id);
		}
		if (this.useItemCache) {
			this.itemCache = cacheFromItems(this.items);
		}
		displayItem(this.id,this.displayed,false);
		checkForChanges(this);
	}

	updateQueue.completed(this);
	updateQueue.update();
	this.saveItems();

}

Delivery.prototype.hasItems = function () {
	var item = this.items[0];
	if (!item) return false;
	if (!item.label) return false;
	return true;
}

Delivery.prototype.updatePackageNames = function () {
	var items = this.items;
	var count = items.length;
	for (var x = 0; x < count; x++) {
		var item = items[x]
		var label = item.label;
		var title = '';
		if (this.name) {
			title = htmlEntities(this.name);
			if (label) title += ': '+label;
		} else if (label) {
			title = label;
		} else if (this.no) {
			title = htmlEntities(this.no);
		} else if (this.domain) {
			title = this.domain.replace(/^(www|m)\./i,'');
		}
		item.title = title;
	}
}

// #mark Delivery items and events

function DeliveryPackage () {

	this.title = translate('Unknown Item');
	this.status = '';
	this.statusDateString = '';
	this.dateString = '';
	this.icon = '';
	this.daysNo = false;
	this.url = '';
	this.isComplete = false;

	this.details = {};
	this.details.dateString = '';
	this.details.location = '';
	this.details.mapLocation = '';
	this.details.status = '';

	this.shippedDate = null;
	this.shippedText = ''; 
	this.shippedBold = false;
	this.shippedExtra = '';
	this.shippedDateString = '';

	this.deliveredDate = null;
	this.deliveredText = '';
	this.deliveredBold = false;
	this.deliveredExtra = '';
	this.deliveredDateString = '';
	this.clickable = false;

}

function DeliveryEvent () {

	this.dateString = null;
	this.status = null;
	this.location = null;	

}

function errorItem (delivery) {
	var item = new DeliveryPackage();
	item.icon = 'error';
	if (delivery.name) {
		item.title = htmlEntities(delivery.name);
	} else if (delivery.no) {
		item.title = htmlEntities(delivery.no);
	} else if (delivery.domain) {
		var title = delivery.domain.replace(/^(www|m)\./i,'');
		item.title = title;
	}
	item.shippedBold = true;
	item.deliveredBold = false;
	return item;
}

// #mark Various utilities

function translate (key) {
	try {
		var ret = translation[key];
		if (ret === undefined || ret === '') ret = key;
		return ret;
	} catch (ex) { }
	return key;
}

function translateFormat (key,value1,value2,value3) {
	key = translate(key);
	if (value1) key = key.replace('%@',value1);
	if (value2) key = key.replace('%@',value2);
	if (value3) key = key.replace('%@',value3);
	return key;
}

function preferredLanguage () {
	var userLang = translate('xx');
	var count = arguments.length;
	if (count == 0) return userLang;
	for (var i = 0; i < count; i++) {
		if (arguments[i] == userLang) return userLang;
	}
	return arguments[0];
}

function preferredLocale () {
	var lang = translate('xx');
	var count = arguments.length;
	for (var x = 0; x < count; x++) {
		var availLoc = arguments[x];
		var availLang = availLoc.substring(0,2);
		if (availLang == lang) return availLoc;
	}
	return arguments[0];
}

function remapString (string, map) {
	if (typeof string !== 'string') return null;
	return map[string] || string
}

function getUnixTime () {
	var currentTime = new Date();
	return currentTime.getTime()/1000;
}

function secondsSince (date) {
	return ((new Date()).getTime() - date.getTime()) / 1000
}

function extraInformationForService (service, type) {
	var info;
	if (typeof extraInformation !== 'undefined') {
		var serviceInfo = extraInformation[service];
		if (serviceInfo) info = serviceInfo[type];
	}
	if (!info && typeof requestExtraInformation === 'function') {
		info = requestExtraInformation(service.from,type);
		if (!extraInformation[service]) extraInformation[service] = [];
		extraInformation[service][type] = info;
	}
	return info;
}

function overlappingMatches (regExText, regExPattern) {
	var regEx = new RegExp(regExPattern,'gim');
	var matches = [];
	var match;
	while (match = regEx.exec(regExText)) {
		matches.push(match[0]);
		regEx.lastIndex = match.index+1;
	}
	return (matches.length > 0) ? matches : null;
}

function stringFromMatch (string, pattern) {
	var result = '';
	if (string) {
		var regex = new RegExp(pattern);
		var matches = string.match(regex);
		if (matches) result = matches[1];
	}
	return result;
}

String.prototype.hashCode = function(){
	var hash = 0, i, l, charCode;
	if (this.length == 0) return hash;
	for (i = 0, l = this.length; i < l; i++) {
		charCode  = this.charCodeAt(i);
		hash  = ((hash<<5)-hash)+charCode;
		hash |= 0; // Convert to 32bit integer
	}
	return hash;
};

// #mark Strings and display text

function getFromName (from,translateName,dropDown) {

	switch(from) {

		case 'www.adobe.com':
			var name = 'Adobe';
			break;
		case 'www.apple.com':
		case 'www.apple.com/zip':
			var name = 'Apple';
			break;
		case 'www.apple.com/japan':
		case 'www.apple.com/japanzip':
			var name = 'Apple';
			if (dropDown) name += ' – '+translate('Japan');
			break;
		case 'www.apple.com/hk':
			var name = 'Apple';
			if (dropDown) name += ' – '+translate('Hong Kong');
			break;
		case 'www.aramex.com':
			var name = 'Aramex';
			if (dropDown) name += ' – '+translate('Shop and Ship');
			break;
		case 'auspost.com.au':
			var name = 'Australia Post';
			break;
		case 'em.canadapost.ca':
			var name = (translateName) ? translate('Canada Post') : 'Canada Post';
			break;
		case 'www.chronopost.fr':
			var name = 'Chronopost';
			if (dropDown) name += ' – '+translate('France');
			break;
		case 'www.dhl.com':
		case 'track.dhl-usa.com':
			var name = 'DHL';
			if (dropDown) name += ' – Express';
			break;
		case 'www.dhlglobalmail.com':
			var name = 'DHL';
			if (dropDown) name += ' – '+translate('US')+' GlobalMail';
			break;
		case 'nolp.dhl.de':
		case 'nolp.dhl.de/int':
			var name = 'DHL';
			if (dropDown) name += ' – '+translate('Germany');
			break;
		case 'www.dpd.com':
			var name = 'DPD';
			if (dropDown) name += ' – '+translate('Germany');
			break;
		case 'www.dpd.co.uk':
			var name = 'DPD';
			if (dropDown) name += ' – '+translate('UK');
			break;
		case 'www.fedex.com':
		case 'spportal.fedex.com':
			var name = 'FedEx';
			break;
		case 'www.gls-germany.com':
			var name = 'GLS';
			break;
		case 'checkout.google.com':
			var name = 'Google Wallet';
			break;
		case 'www.hlg.de':
			var name = 'Hermes Logistik Gruppe';
			if (dropDown) name += ' – '+translate('Germany');
			break;
		case 'www.hongkongpost.com':
			var name = 'Hongkong Post';
			break;
		case 'www.japanpost.jp/ems':
		case 'www.japanpost.jp/m10':
		case 'www.japanpost.jp/reg':
		case 'www.japanpost.jp/yupack':
			var name = (translateName) ? translate('Japan Post') : 'Japan Post';
			break;
		case 'www.laposte.fr':
			var name = 'La Poste';
			if (dropDown) name += ' – '+translate('France');
			break;
		case 'www.lasership.com':
			var name = 'LaserShip';
			break;
		case 'www.ontrac.com':
			var name = 'OnTrac';
			break;
		case 'www.parcelforce.com':
			var name = 'Parcelforce';
			break;
		case 'www.post.at':
			var name = (translateName)?'Post':'Post.at';
			if (dropDown) name += ' – '+translate('Austria');
			break;
		case 'www.postdanmark.dk':
			var name = 'PostNord';
			if (dropDown) name += ' – '+translate('Denmark');
			break;
		case 'www.poste.it':
		case 'www.poste.it/posta1':
		case 'www.poste.it/pacco1':
		case 'www.poste.it/maxi':
		case 'www.poste.it/pacco':
		case 'www.poste.it/ems':
			var name = (translateName) ? 'Poste Italiane' : 'Poste.it';
			break;
		case 'www.posten.no':
			var name = (translateName) ? 'Posten' : 'Posten.no';
			if (dropDown) name += ' – '+translate('Norway');
			break;
		case 'www.posten.se': // www.postnord.se
			var name = 'PostNord';
			if (dropDown) name += ' – '+translate('Sweden');
			break;
		case 'www.postnl.nl':
		case 'www.tntpost.nl':
			var name = 'PostNL';
			break;
		case 'www.purolator.com':
			var name = 'Purolator';
			break;
		case 'www.royalmail.com':
			var name = 'Royal Mail';
			break;
		case 'www.sagawa-exp.co.jp':
			var name = (translateName) ? translate('Sagawa Express') : 'Sagawa';
			break;
		case 'www.post.ch':
			var name = (translateName) ? translate('Swiss Post') : 'Swiss Post';
			break;
		case 'www.tnt.com':
			var name = 'TNT';
			if (dropDown) name += ' – '+translate('by consignment number');
			break;
		case 'www.tnt.com/ref':
			var name = 'TNT';
			if (dropDown) name += ' – '+translate('by reference number');
			break;
		case 'www.tntexpress.com.au':
			var name = 'TNT Express';
			if (dropDown) name += ' – '+translate('Australia');
			break;
		case 'www.ups.com':
			var name = 'UPS';
			break;
		case 'www.ups-mi.net':
			var name = 'UPS Mail Innovations';
			break;
		case 'www.usps.com':
			var name = 'U.S. Postal Service';
			break;
		case 'www.kuronekoyamato.co.jp':
			var name = (translateName) ? translate('Yamato Transport') : 'Yamato Transport';
			break;
		case 'www.yodel.co.uk':
			var name = 'Yodel';
			break;
		case 'www.amazon.com':
		case 'www.amazon.ca':
		case 'www.amazon.co.uk':
		case 'www.amazon.co.jp':
		case 'www.amazon.de':
		case 'www.amazon.at':
		case 'www.amazon.es':
		case 'www.amazon.fr':
		case 'www.amazon.in':
		case 'www.amazon.it':
			var name = 'A'+from.substring(5);
			break;
		case 'www.amazon.com/url':
			var name = 'Amazon';
			if (dropDown) name += ' – '+translate('by tracking link');
			break;
		case 'other':
			var name = (translateName) ? translate('Other') : 'Other';
			break;
		default:
			var name = from;
			break;

	}
	return name;

}

function stripWhiteSpace (string) {
	// \s should match [\f\n\r\t\v\u00A0\u2028\u2029] but \u2028 is needed for iOS 4
	return (typeof string == 'string') ? string.replace(/[\s\n\u2028]{2,}|[\n\u2028]+/g,' ') : '';
}

function trim (string) {
	if (typeof string == 'string') {
		// Remove non-breaking space characters (&#160; or \xAO)
		// while (string.indexOf(' ') > -1) string = string.replace(' ',' ');
		string = string.replace(/^[\s]+|[\s]+$/g,'');
	}
	return string;
}

function cleanSpaces (string) {
	return (typeof string == 'string') ? string.replace(/(?:\s|&nbsp;){2,}|&nbsp;/g,' ') : '';
}

function clearEmptyString (string) {
	return (typeof string == 'string') ? string.replace(/^ ?-+ ?$/,'') : '';
}

function stripHTML (string) {
	if (typeof string == 'string') {

		// Deal with some really ugly HTML from UPS Mail Innovations
		string = string.replace(/<a [^>]*onmouseover="([^"]*)"[^>]*>/g,' ');

		// Remove script, noscript, and style tags
		string = string.replace(/<(noscript|script|style) ?[^>]*>(.*?)<\/\1>/ig,' ');

		// Remove HTML tags
		string = string.replace(/<[^>]*>/g,' ');

		// Remove non-breaking space characters (&#160; or \xAO) and reduce mutiple spaces
		// while (string.indexOf(' ') > -1) string = string.replace(' ',' ');
		string = cleanSpaces(string);

		// Convert some HTML entities
		string = string.replace(/&quot;/g,'"');

	}
	return trim(string);
}

function htmlEntities (string) {
	if (typeof string == 'string') {
		string = string.replace(/\&/g,'&#38;');
		string = string.replace(/"/g,'&quot;');
		string = string.replace(/</g,'&#60;');
		string = string.replace(/>/g,'&#62;');
	}
	return string;
}

function decodeDecimalEntities (string) {
	return string.replace(/&#(\d+);/g, 
		function (whole,parens) {
			if (parens >= -32768 && parens <= 65535) {
				return String.fromCharCode(parens);
			} else return whole;
		});
}

function removeHtmlEntities (string) {
	if (typeof string == 'string') {
		var entities = {
			'lt':'<',
			'gt':'>',
			'amp':'&',
			'apos':'\'',
			'quot':'"',
			'lsquo':'‘',
			'rsquo':'’',
			'ldquo':'“',
			'rdquo':'”',
			'nbsp':' ',
			'ndash':'–',
			'mdash':'—',
			'auml':'ä',
			'euml':'ë',
			'iuml':'ï',
			'ouml':'ö',
			'uuml':'ü',
			'Auml':'Ä',
			'Euml':'Ë',
			'Iuml':'Ï',
			'Ouml':'Ö',
			'Uuml':'Ü',
			'szlig':'ß',
			'aacute':'á',
			'eacute':'é',
			'iacute':'í',
			'oacute':'ó',
			'uacute':'ú',
			'Aacute':'Á',
			'Eacute':'É',
			'Iacute':'Í',
			'Oacute':'Ó',
			'Uacute':'Ú',
			'agrave':'à',
			'egrave':'è',
			'igrave':'ì',
			'ograve':'ò',
			'ugrave':'ù',
			'Agrave':'À',
			'Egrave':'È',
			'Igrave':'Ì',
			'Ograve':'Ò',
			'Ugrave':'Ù',
			'acirc':'â',
			'ecirc':'ê',
			'icirc':'î',
			'ocirc':'ô',
			'ucirc':'û',
			'Acirc':'Â',
			'Ecirc':'Ê',
			'Icirc':'Î',
			'Ocirc':'Ô',
			'Ucirc':'Û',
			'agrave':'à',
			'egrave':'è',
			'igrave':'ì',
			'ograve':'ò',
			'ugrave':'ù',
			'Agrave':'À',
			'Egrave':'È',
			'Igrave':'Ì',
			'Ograve':'Ò',
			'Ugrave':'Ù',
			'ntilde':'ñ',
			'atilde':'ã',
			'otilde':'õ',
			'Ntilde':'Ñ',
			'Atilde':'Ã',
			'Otilde':'Õ',
			'aring':'å',
			'Aring':'Å',
			'aelig':'æ',
			'AElig':'Æ',
			'oslash':'ø',
			'Oslash':'Ø',
			'copy':'©',
			'trade':'™',
			'reg':'®',
			'bull':'•',
			'deg':'°',
			'quest':'?',
			'iquest':'¿',
			'frac14':'¼',
			'frac12':'½',
			'frac34':'¾'
		};
		return string.replace(/&(#(?:x[0-9a-f]+|\d+)|[a-z]+[0-9]*);?/gi, function(whole,parens) {
			if (parens[0] === "#") {
				var charCode = parens[1].toLowerCase() === 'x' ? parseInt(parens.substr(2),16) : parseInt(parens.substr(1),10);
				return String.fromCharCode(charCode);
			} else return entities.hasOwnProperty(parens) ? entities[parens] : whole;
		});
	}
	return string;
}

function doublePlusEncode (string) {
	return encodeURIComponent(string).replace(/%2B/g,'%252B');
}

// #mark JSON parsing

function parseJSONInHTML (html, pattern) {
	if (!html) return null;
	var result = html.match(pattern);
	if (!result) return null;
	return parseJSON(result[1]);
}

function parseJSON (json) {
	var object = null;
	try {
		object = JSON.parse(json);
	} catch (err) {
		logMessage(1,err);
	}
	return object;
}

// #mark Date parsing

function DeliveryDate (year, month, day, hours, minutes, seconds, timeZone, lang, future) {
	this.day = parseInt(day, 10);
	if (lang) {
		this.month = monthStringToInt(month, lang);
	} else {
		this.month = parseInt(month, 10);
	}
	if (year) {
		if (year.length == 2) {
			var currentYear = new Date().getFullYear();
			year = currentYear.toString().slice(0,-2) + year;
		}
		this.year = year;
	} else {
		this.guessYear(future);
	}
	if (timeZone) {
		this.hours = parseInt(hours, 10);
		this.minutes = parseInt(minutes, 10);
		this.seconds = parseInt(seconds, 10);
		this.timeZone = timeZone;
	}
}

DeliveryDate.prototype.guessYear = function (future) {

	var today = new Date();
	var currentYear = today.getFullYear();
	var currentMonth = today.getMonth();
	var currentDay = today.getDate();
	today = new Date(currentYear, currentMonth, currentDay); // Remove h/m/s
	var recentDays = (future) ? 15 : 180;
	var recentPast = new Date(currentYear, currentMonth, currentDay - recentDays);

	var month = this.month - 1;
	var day = this.day;
	var year = currentYear;

	// If using the previous year puts it in the recent past, assume that's correct
	var guess = new Date(year - 1, month, day);
	if (guess > recentPast) year -= 1;

	// If it's probably a future date, and the current guess is not in the recent 
	// past, or this month, then add a year
	if (future) {
		var guess = new Date(year, month, day);
		if (guess < recentPast && guess.getMonth() != currentMonth) year += 1;
	}

	this.year = year;

}

DeliveryDate.prototype.isValid = function () {
	if (this.year === null) return false;
	if (this.month === null) return false;
	if (this.day === null) return false;
	return true;
}

DeliveryDate.prototype.pad =  function (number) {
	if (number < 10) return '0'+number;
	return number;
}

DeliveryDate.prototype.toString = function (separator) {
	return this.toStringUsingSeparator('-');
}

DeliveryDate.prototype.toStringUsingSeparator = function (separator) {
	if (!this.isValid()) return null;
	var year = this.year;
	var month = this.pad(this.month);
	var day = this.pad(this.day);
	return year + separator + month + separator + day;
}

DeliveryDate.prototype.toISOString = function () {
	var string = this.toString();
	var timeZone = this.timeZone;
	if (!timeZone) return string;
	var hours = this.pad(this.hours);
	var minutes = this.pad(this.minutes);
	var seconds = this.pad(this.seconds);
	return string + 'T' + hours + ':' + minutes + ':' + seconds + timeZone;
}

DeliveryDate.prototype.toDate = function () {
	if (!this.isValid()) return null;
	if (this.timeZone) {
		return new Date(this.year, this.month-1, this.day, this.hours, this.minutes, this.seconds);
	}
	return new Date(this.year, this.month-1, this.day);
}

DeliveryDate.prototype.toTimeStamp = function () {
	return this.valueOf();
}

DeliveryDate.prototype.valueOf = function () {
	var date = this.toDate();
	if (!date) return null;
	return date.valueOf();
}

function dateFromISOString (string) {
	if (typeof string != 'string') return null;
	var match = string.match(/^([0-9]{4})-([0-9]{2})-([0-9]{2})T([0-9]{2}):([0-9]{2}):([0-9]{2})(Z|[-+][0-9]{2}:[0-9]{2})$/);
	if (match) return new DeliveryDate(match[1], match[2], match[3], match[4], match[5], match[6], match[7]);
	var match = string.match(/^([0-9]{4})-([0-9]{2})-([0-9]{2})(?:[^0-9]|$)/);
	if (match) return new DeliveryDate(match[1], match[2], match[3]);
	return null;
}

function timeStampToISOString (timeStamp) {
	if (!timeStamp) return null;
	var date = new Date(timeStamp);
	return date.toISOString();
}

function dateFromParts (year, month, day, lang, future) {
	return new DeliveryDate(year, month, day, null, null, null, null, lang, future);
}

function monthStringToInt (string, lang) {
	var month = string.replace(/[^a-z\u00E0-\u00FC]/i, '').toLowerCase();
	switch (lang) {

		case 'en':
			var prefix = month.substring(0, 3);
			switch (prefix) {
				case 'jan': return 1;
				case 'feb': return 2;
				case 'mar': return 3;
				case 'apr': return 4;
				case 'may': return 5;
				case 'jun': return 6;
				case 'jul': return 7;
				case 'aug': return 8;
				case 'sep': return 9;
				case 'oct': return 10;
				case 'nov': return 11;
				case 'dec': return 12;
			}
			return null;

		case 'de':
			var prefix = month.substring(0, 3);
			switch (prefix) {
				case 'jan': case 'jän': return 1;
				case 'feb': return 2;
				case 'mär': return 3;
				case 'apr': return 4;
				case 'mai': return 5;
				case 'jun': return 6;
				case 'jul': return 7;
				case 'aug': return 8;
				case 'sep': return 9;
				case 'okt': return 10;
				case 'nov': return 11;
				case 'dez': return 12;
			}
			return null;

		case 'es':
			var prefix = month.substring(0, 3);
			switch (prefix) {
				case 'ene': return 1;
				case 'feb': return 2;
				case 'mar': return 3;
				case 'abr': return 4;
				case 'may': return 5;
				case 'jun': return 6;
				case 'jul': return 7;
				case 'ago': return 8;
				case 'sep': case 'set': return 9;
				case 'oct': return 10;
				case 'nov': return 11;
				case 'dic': return 12;
			}
			return null;

		case 'fr':
			var prefix = month.substring(0, 4);
			switch (prefix) {
				case 'janv': return 1;
				case 'févr': case 'fevr': return 2;
				case 'mars': return 3;
				case 'avri': case 'avr': return 4;
				case 'mai': return 5;
				case 'juin': return 6;
				case 'juil': return 7;
				case 'août': case 'aout': return 8;
				case 'sept': return 9;
				case 'octo': case 'oct': return 10;
				case 'nove': case 'nov': return 11;
				case 'déce': case 'dece': case 'déc': case 'dec': return 12;
			}
			return null;

		case 'it':
			var prefix = month.substring(0, 3);
			switch (prefix) {
				case 'gen': return 1;
				case 'feb': return 2;
				case 'mar': return 3;
				case 'apr': return 4;
				case 'mag': return 5;
				case 'giu': return 6;
				case 'lug': return 7;
				case 'ago': case 'ag': return 8;
				case 'set': return 9;
				case 'ott': return 10;
				case 'nov': return 11;
				case 'dic': return 12;
			}
			return null;



		default:
			return null;
	}
}

function possibleWeekDay (string) {
	var regEx = /^[a-z\u00E0-\u00FC']+$/i;
	return regEx.test(string);
}

function dateFromWeekDay (string, language, referenceDate) {
	// Note: referenceDate is not currently used
	if (typeof string != 'string') return null;
	var offset = false;
	var weekday = false;
	var lowercaseString = string.toLowerCase();
	switch (language) {
		case 'de':
			switch (lowercaseString) {
				case 'heute':		offset = 0; break;
				case 'morgen':		offset = 1; break;
				case 'gestern':		offset = -1; break;
				case 'sonntag':		weekday = 0; break;
				case 'montag':		weekday = 1; break;
				case 'dienstag':	weekday = 2; break;
				case 'mittwoch':	weekday = 3; break;
				case 'donnerstag':	weekday = 4; break;
				case 'freitag':		weekday = 5; break;
				case 'sonnabend':	weekday = 6; break;
			}
			break;

		case 'es':
			switch (lowercaseString) {
				case 'hoy':			offset = 0; break;
				case 'mañana':		offset = 1; break;
				case 'ayer':		offset = -1; break;
				case 'domingo':		weekday = 0; break;
				case 'lunes':		weekday = 1; break;
				case 'martes':		weekday = 2; break;
				case 'miércoles':	weekday = 3; break;
				case 'jueves':		weekday = 4; break;
				case 'viernes':		weekday = 5; break;
				case 'sábado':		weekday = 6; break;
			}
			break;

		case 'fr':
			switch (lowercaseString) {
				case 'aujourd\'hui':offset = 0; break;
				case 'demain':		offset = 1; break;
				case 'hier':		offset = -1; break;
				case 'dimanche':	weekday = 0; break;
				case 'lundi':		weekday = 1; break;
				case 'mardi':		weekday = 2; break;
				case 'mercredi':	weekday = 3; break;
				case 'jeudi':		weekday = 4; break;
				case 'vendredi':	weekday = 5; break;
				case 'samedi':		weekday = 6; break;
			}
			break;
		case 'it':
			switch (lowercaseString) {
				case 'oggi':		offset = 0; break;
				case 'domani':		offset = 1; break;
				case 'ieri':		offset = -1; break;
				case 'domenica':	weekday = 0; break;
				case 'lunedì':		weekday = 1; break;
				case 'martedì':		weekday = 2; break;
				case 'mercoledì':	weekday = 3; break;
				case 'giovedì':		weekday = 4; break;
				case 'venerdì':		weekday = 5; break;
				case 'sabato':		weekday = 6; break;
			}
			break;
		case 'en':
		default:
			switch (lowercaseString) {
				case 'today':		offset = 0; break;
				case 'tomorrow':	offset = 1; break;
				case 'yesterday':	offset = -1; break;
				case 'sunday':		weekday = 0; break;
				case 'monday':		weekday = 1; break;
				case 'tuesday':		weekday = 2; break;
				case 'wednesday':	weekday = 3; break;
				case 'thursday':	weekday = 4; break;
				case 'friday':		weekday = 5; break;
				case 'saturday':	weekday = 6; break;
			}
			break;
	}
	var date = null;
	if (offset !== false) {
		date = (referenceDate) ? new Date(referenceDate) : new Date();
		date.setDate(date.getDate() + offset);
	} else if (weekday !== false) {
		date = (referenceDate) ? new Date(referenceDate) : new Date();
		var day = date.getDay();
		if (weekday <= day) weekday += 7;
		weekday -= day;
		date.setDate(date.getDate() + weekday);
	}
	if (!date) return null;
	return new DeliveryDate(date.getFullYear(), date.getMonth()+1, date.getDate());
}

function daysUntilDate (timeStamp) {
	if (dashboardWidget && DeliveryStatus) {
		return DeliveryStatus.daysUntilDate(timeStamp / 1000);
	} else {
		// Note: this is approximate
		var date = new Date();
		date.setTime(timeStamp);
		date.setHours(0);
		date.setMinutes(0);
		date.setSeconds(0);
		return Math.ceil((date - new Date()) / 86400000);
	}
}

function monthsUntilDate (fromDate) {
	if (dashboardWidget && DeliveryStatus) {
		return DeliveryStatus.monthsUntilDate(fromDate / 1000);
	}
}

function ParseOptions (from) {
	this.from = from;
	this.locale = null;
}

// #mark Logging

function logRequest (level,request) {
	if (typeof logLevel === 'undefined') return;
	if (logLevel >= level && request) {
		var message = '';
		if (request.url) message += request.url+'\n';
		message += 'Status: '+request.status+'\n';
		message += 'Ready State: '+request.readyState+'\n';
		if (request.getAllResponseHeaders) message += request.getAllResponseHeaders()+'\n';
		if (message) message += '\n';
		alert(message+request.responseText);
	}
}

function logMessage (level,message) {
	if (typeof logLevel === 'undefined') return;
	if (logLevel >= level) alert(message);
}
